/*
 * var.c - Variable handling
 *
 * Written 2001,2002 by Werner Almesberger
 * Copyright 2001 EPFL-ICA, Network Robots
 * Copyright 2002 Bivio Networks, Network Robots
 */


#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "config.h"
#include "util.h"
#include "error.h"
#include "location.h"
#include "var.h"


static VAR *variables = NULL;
static VAR **last = &variables;
static int level = 0;


/* ----- Variable lookup and creation -------------------------------------- */


static VAR *var_new(const char *name)
{
    VAR *var;

    var = alloc_t(VAR);
    var->name = stralloc(name);
    var->data = data_none();
    var->data_ref = NULL;
    var->level = -1;
    var->level = level;
    var->usage = vu_new;
    var->location = current_location();
    var->next = NULL;
    *last = var;
    last = &var->next;
    return var;
}


VAR *var_find(const char *name)
{
    VAR **walk,**var;

    var = NULL;
    for (walk = &variables; *walk; walk = &(*walk)->next)
	if (!strcmp((*walk)->name,name)) var = walk;
    if (!var) var = walk;
    if (!*var) *var = var_new(name);
    (*var)->found = current_location();
    debugf("var_find: %s -> %p (%s)",name,*var,
      (*var)->usage == vu_new ? "new" : (*var)->usage == vu_unused ? "unused"
      : (*var)->usage == vu_used ? "used" : "forward");
    return *var;
}


/* ----- Variable scopes --------------------------------------------------- */


void var_begin_scope(void)
{
    level++;
    debugf("var_begin_scope (now %d)",level);
}


void var_end_scope(void)
{
    debugf("var_end_scope(was %d)",level);
    for (last = &variables; *last; last= &(*last)->next)
	if ((*last)->level == level) break;
    while (*last) {
	VAR *this;

	this = *last;
	*last = this->next;
	if (this->usage == vu_unused && warn_unused)
	    lwarnf(this->location,"unused variable %s",this->name);
	if (this->usage == vu_forward)
	    lerrorf(this->location,"unresolved forward-declared variable %s",
	      this->name);
	/*
	 * Don't destroy object - we need it for output
	 */
	debugf("  removing %s",this->name);
	free((char *) this->name);
	free(this);
    }
    level--;
}


void var_outer_scope(void)
{
    debugf("var_outer_scope(was %d)",level);
    level--;
}


/* ----- Variable access --------------------------------------------------- */


static struct var_access {
    VAR *var;
    const char *new_entry;
    DATA clone;			/* cloned data, so that we don't clobber orig
				   e.g. if in outer scope */
    DATA *data;			/* insertion point */
    const DATA *orig_data;	/* retrieval point */
    const DATA_ASSOC *entry;	/* position in associative array */
    struct var_access *next;	/* next on stack */
} current_access;

/*
 * Write accesses can be nested using compound expressions, so we need
 * to stack them. Read accesses are never nested, so a static top of
 * stack element is sufficient.
 */


void var_begin_access(VAR *var)
{
    current_access.var = var;
    current_access.new_entry = NULL;
    current_access.clone = data_clone(var->data);
    current_access.data = &current_access.clone;
    current_access.orig_data = &var->data;
    current_access.entry = NULL;
}


void var_push_access(void)
{
    struct var_access *new;

    new = alloc_t(struct var_access);
    *new = current_access;
    current_access.next = new;
}


void var_pop_access(void)
{
    struct var_access *top;

    top = current_access.next;
    current_access = *top;
    free(top);
}


void var_follow(const char *name)
{
    DATA_ASSOC **entry;

    if (current_access.data->type != dt_assoc) {
	data_destroy(*current_access.data);
	*current_access.data = data_assoc();
    }
    for (entry = &current_access.data->u.assoc; *entry;
      entry = &(*entry)->next)
        if (!strcmp((*entry)->name,name)) break;
    if (*entry) {
	const DATA_ASSOC *walk;

	for (walk = current_access.orig_data->u.assoc; walk; walk = walk->next)
	    if (!strcmp(walk->name,name)) break;
	assert(walk);
	current_access.orig_data = &walk->data;
    }
    else {
	*entry = alloc_t(DATA_ASSOC);
	(*entry)->name = stralloc(name);
	(*entry)->data = data_none();
	(*entry)->parent = current_access.entry;
	(*entry)->next = NULL;
	if (!current_access.new_entry)
	    current_access.new_entry = (*entry)->name;
	current_access.orig_data = NULL;
    }
    current_access.data = &(*entry)->data;
    current_access.entry = *entry;
}


void var_set(DATA data)
{
    debugf("var_set %s",current_access.var->name);
    register_var(current_access.var->name,current_access.entry,data);
    if (current_access.var->usage == vu_forward) {
	if (current_access.entry)
	    yyerror("cannot access structure entries in forward-declared "
	      "variable");
	current_access.var->usage = vu_used;
	*current_access.data = *current_access.var->data_ref = data;
	current_access.var->data = current_access.clone;
	return;
    }
    if (!current_access.entry || current_access.var->data.type != dt_assoc) {
	if (current_access.var->usage != vu_new &&
	 current_access.var->level == level) {
	    if (current_access.var->usage == vu_unused && warn_unused)
		lwarnf(current_access.var->location,"unused variable %s",
		  current_access.var->name);
	    if (warn_redefine)
		lwarnf(current_access.var->found,"variable %s redefined",
		  current_access.var->name);
	    current_access.var->location = current_access.var->found;
	    if (!var_use_file) data_destroy(current_access.var->data);
	}
	if (current_access.var->level != level)
	    current_access.var = var_new(current_access.var->name);
	      /* no need to reset current_access.data here, because it can only
		 point to &current_access.clone */
	current_access.var->usage = vu_unused;
    }
    *current_access.data = data;
    current_access.var->data = current_access.clone;
}


DATA *var_forward(void)
{
    debugf("var_forward %s",current_access.var->name);
    if (current_access.var->usage != vu_new &&
      current_access.var->usage != vu_forward &&
      current_access.var->level == level)
	yyerrorf("variable %s redefined",current_access.var->name);
    if (current_access.entry)
	yyerrorf("cannot forward-declare structure entries");
    if (current_access.var->level != level)
	current_access.var = var_new(current_access.var->name);
    if (!current_access.var->data_ref)
	current_access.var->data_ref = alloc_t(DATA);
    current_access.var->usage = vu_forward;
    return current_access.var->data_ref;
}


int var_initialized(const VAR *var)
{
    return !(var->usage == vu_new || var->usage == vu_forward);
}


DATA var_get(void)
{
    debugf("var_get %s",current_access.var->name);
    if (!var_initialized(current_access.var))
	yyerrorf("variable %s is uninitialized",current_access.var->name);
    if (current_access.var->data.type != dt_assoc &&
      current_access.clone.type == dt_assoc)
	yyerrorf("expected structure, not type %s",
	  type_name(current_access.var->data.type));
    if (current_access.new_entry)
	yyerrorf("no entry \"%s\" in structure",current_access.new_entry);
    data_destroy(current_access.clone);
    current_access.var->usage = vu_used;
    return *current_access.orig_data;
}
